@@ -1,8 +1,51 @@
-#!/usr/bin/env python
+# -*- coding: utf-8 -*-
+
+# Copyright OpenStack Foundation and OpenDev Contributors
+#
+# Licensed under the Apache License, Version 2.0 (the "License"); you may
+# not use this file except in compliance with the License.  You may obtain
+# a copy of the License at
+#
+#    http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
+# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
+# License for the specific language governing permissions and limitations
+# under the License.
+
 from __future__ import print_function
 
+import argparse
+import configparser
+import datetime
+import getpass
+import json
+import os
+import re
+import shlex
+import subprocess
+import sys
+import textwrap
+from urllib.parse import urlencode
+from urllib.parse import urljoin
+from urllib.parse import urlparse
+
+import pkg_resources
+import requests
+
+
+VERBOSE = False
+UPDATE = False
+LOCAL_MODE = 'GITREVIEW_LOCAL_MODE' in os.environ
+CONFIGDIR = os.path.expanduser("~/.config/git-review")
+GLOBAL_CONFIG = "/etc/git-review/git-review.conf"
+USER_CONFIG = os.path.join(CONFIGDIR, "git-review.conf")
+DEFAULTS = dict(scheme='ssh', hostname=False, port=None, project=False,
+                branch='master', remote="gerrit", rebase="1",
+                track="0", usepushurl="0", notopic=False, branchauthor="name")
 COPYRIGHT = """\
-Copyright (C) 2011-2012 OpenStack LLC.
+Copyright OpenStack Foundation and OpenDev Contributors
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -16,49 +59,9 @@ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 implied.
 
 See the License for the specific language governing permissions and
-limitations under the License."""
+limitations under the License.
+"""
 
-import argparse
-import datetime
-import getpass
-import json
-import os
-import re
-import shlex
-import subprocess
-import sys
-import textwrap
-
-import pkg_resources
-import requests
-
-if sys.version < '3':
-    import ConfigParser
-    import urllib
-    import urlparse
-    urlencode = urllib.urlencode
-    urljoin = urlparse.urljoin
-    urlparse = urlparse.urlparse
-    do_input = raw_input
-else:
-    import configparser as ConfigParser
-
-    import urllib.parse
-    import urllib.request
-    urlencode = urllib.parse.urlencode
-    urljoin = urllib.parse.urljoin
-    urlparse = urllib.parse.urlparse
-    do_input = input
-
-VERBOSE = False
-UPDATE = False
-LOCAL_MODE = 'GITREVIEW_LOCAL_MODE' in os.environ
-CONFIGDIR = os.path.expanduser("~/.config/git-review")
-GLOBAL_CONFIG = "/etc/git-review/git-review.conf"
-USER_CONFIG = os.path.join(CONFIGDIR, "git-review.conf")
-DEFAULTS = dict(scheme='ssh', hostname=False, port=None, project=False,
-                branch='master', remote="gerrit", rebase="1",
-                track="0", usepushurl="0")
 
 _branch_name = None
 _has_color = None
@@ -66,6 +69,7 @@ _use_color = None
 _orig_head = None
 _rewrites = None
 _rewrites_push = None
+BRANCHAUTHOR = None
 
 
 class colors(object):
@@ -76,7 +80,7 @@ class colors(object):
 
 
 class GitReviewException(Exception):
-    EXIT_CODE = 127
+    EXIT_CODE = 1
 
 
 class CommandFailed(GitReviewException):
@@ -133,11 +137,7 @@ def run_command_status(*argv, **kwargs):
     if VERBOSE:
         print(datetime.datetime.now(), "Running:", " ".join(argv))
     if len(argv) == 1:
-        # for python2 compatibility with shlex
-        if sys.version_info < (3,) and isinstance(argv[0], unicode):
-            argv = shlex.split(argv[0].encode('utf-8'))
-        else:
-            argv = shlex.split(str(argv[0]))
+        argv = shlex.split(str(argv[0]))
     stdin = kwargs.pop('stdin', None)
     newenv = os.environ.copy()
     newenv['LANG'] = 'C'
@@ -163,7 +163,7 @@ def run_command_exc(klazz, *argv, **env):
     klazz should be derived from CommandFailed
     """
     (rc, output) = run_command_status(*argv, **env)
-    if rc != 0:
+    if rc:
         raise klazz(rc, output, argv, env)
     return output
 
@@ -172,10 +172,10 @@ def git_credentials(url):
     """Return credentials using git credential or None."""
     cmd = 'git', 'credential', 'fill'
     stdin = 'url=%s' % url
-    rc, out = run_command_status(*cmd, stdin=stdin.encode('utf-8'))
+    rc, out = run_command_status(*cmd, stdin=stdin)
     if rc:
         return None
-    data = dict(l.split('=', 1) for l in out.splitlines())
+    data = dict(line.split('=', 1) for line in out.splitlines())
     return data['username'], data['password']
 
 
@@ -254,7 +254,7 @@ def run_custom_script(action):
             returns.append((status, output, fpath))
 
     for (status, output, path) in returns:
-        if status is not None and status != 0:
+        if status:
             raise CustomScriptException(status, output, [path], {})
         elif output and VERBOSE:
             print("script %s output is:" % (path))
@@ -272,12 +272,15 @@ def git_config_get_value(section, option, default=None, as_bool=False):
     try:
         result = run_command_exc(GitConfigException, *cmd).strip()
         if VERBOSE:
-            print(datetime.datetime.now(), "result:", result)
+            print(datetime.datetime.now(), "... %s.%s = %s"
+                  % (section, option, result))
         return result
     except GitConfigException as exc:
         if exc.rc == 1:
-            if VERBOSE:
-                print(datetime.datetime.now(), "using default:", default)
+            if VERBOSE and default is not None:
+                print(datetime.datetime.now(),
+                      "... nothing in git config, returning func parameter:",
+                      default)
             return default
         raise
 
@@ -300,9 +303,16 @@ class Config(object):
                 self.config.update(load_config_file(filename))
 
     def __getitem__(self, key):
+        """Let 'git config --get' override every Config['key'] access"""
         value = git_config_get_value('gitreview', key)
         if value is None:
             value = self.config[key]
+        # "--verbose" doesn't trace *early* invocations; for that you
+        # must change the value at the top of this file (*and* pass
+        # --verbose)
+        if VERBOSE:
+            print(datetime.datetime.now(),
+                  "Config['%s'] = %s " % (key, value))
         return value
 
 
@@ -338,10 +348,10 @@ def set_hooks_commit_msg(remote, target_file):
         else:
             (hostname, username, port, project_name) = \
                 parse_gerrit_ssh_params_from_git_url(remote_url)
-            if username is None:
-                userhost = hostname
-            else:
+            if username:
                 userhost = "%s@%s" % (username, hostname)
+            else:
+                userhost = hostname
             # OS independent target file
             scp_target_file = target_file.replace(os.sep, "/")
             cmd = ["scp", userhost + ":hooks/commit-msg", scp_target_file]
@@ -353,6 +363,12 @@ def set_hooks_commit_msg(remote, target_file):
                            % (userhost, (":%s" % port) if port else "")
                 print("Fetching commit hook from: %s" % hook_url)
             run_command_exc(CannotInstallHook, *cmd)
+        # If there are submodules, the hook needs to be installed into
+        # each of them.
+        run_command_exc(
+            CannotInstallHook,
+            "git", "submodule", "foreach",
+            'cp -p %s "$(git rev-parse --git-dir)/hooks/"' % target_file)
 
     if not os.access(target_file, os.X_OK):
         os.chmod(target_file, os.path.stat.S_IREAD | os.path.stat.S_IEXEC)
@@ -367,9 +383,7 @@ def test_remote_url(remote_url):
             print("%s worked. Description: %s" % (remote_url, description))
         return True
     else:
-        if VERBOSE:
-            print("%s did not work. Description: %s" % (
-                remote_url, description))
+        print("%s did not work. Description: %s" % (remote_url, description))
         return False
 
 
@@ -378,10 +392,10 @@ def make_remote_url(scheme, username, hostname, port, project):
     if port is None and scheme == 'ssh':
         port = 29418
     hostport = '%s:%s' % (hostname, port) if port else hostname
-    if username is None:
-        return "%s://%s/%s" % (scheme, hostport, project)
-    else:
+    if username:
         return "%s://%s@%s/%s" % (scheme, username, hostport, project)
+    else:
+        return "%s://%s/%s" % (scheme, hostport, project)
 
 
 def add_remote(scheme, hostname, port, project, remote, usepushurl):
@@ -397,7 +411,7 @@ def add_remote(scheme, hostname, port, project, remote, usepushurl):
         print("No remote set, testing %s" % remote_url)
     if not test_remote_url(remote_url):
         print("Could not connect to gerrit.")
-        username = do_input("Enter your gerrit username: ")
+        username = input("Enter your gerrit username: ")
         remote_url = make_remote_url(scheme, username, hostname, port, project)
         print("Trying again with %s" % remote_url)
         if not test_remote_url(remote_url):
@@ -415,7 +429,7 @@ def add_remote(scheme, hostname, port, project, remote, usepushurl):
     print("\t%s" % remote_url)
 
     (status, remote_output) = run_command_status(cmd)
-    if status != 0:
+    if status:
         raise CommandFailed(status, remote_output, cmd, {})
 
     if asked_for_username:
@@ -472,8 +486,8 @@ def alias_url(url, rewrite_push):
         # there is a longer insteadOf alias.
         longest = None
         for alias in rewrites:
-            if (url.startswith(alias)
-                    and (longest is None or len(longest) < len(alias))):
+            if url.startswith(alias) and (
+                    longest is None or len(longest) < len(alias)):
                 longest = alias
 
         if longest:
@@ -528,7 +542,7 @@ def parse_gerrit_ssh_params_from_git_url(git_url):
             hostname = parsed_url.path[2:].split("/")[0]
 
         if "@" in hostname:
-            (username, hostname) = hostname.split("@")
+            (username, _, hostname) = hostname.rpartition("@")
         if ":" in hostname:
             (hostname, port) = hostname.split(":")
 
@@ -706,7 +720,7 @@ def check_color_support():
 
 def load_config_file(config_file):
     """Load configuration options from a file."""
-    configParser = ConfigParser.ConfigParser()
+    configParser = configparser.ConfigParser()
     configParser.read(config_file)
     options = {
         'scheme': 'scheme',
@@ -717,7 +731,9 @@ def load_config_file(config_file):
         'remote': 'defaultremote',
         'rebase': 'defaultrebase',
         'track': 'track',
+        'notopic': 'notopic',
         'usepushurl': 'usepushurl',
+        'branchauthor': 'branchauthor',
     }
     config = {}
     for config_key, option_name in options.items():
@@ -808,8 +824,8 @@ def check_remote(branch, remote, scheme, hostname, port, project,
             printwrap("Please set the push-url on your origin remote to the "
                       "location of your gerrit server and try again")
         else:
-            printwrap("Please manually create a remote "
-                      "named \"%s\" and try again." % remote)
+            printwrap("Please manually create a remote named \"%s\" or "
+                      "rename the default one and try again." % remote)
         sys.exit(1)
 
     # Gerrit remote not present, try to add it
@@ -846,6 +862,23 @@ def rebase_changes(branch, remote, interactive=True):
         return False
     _orig_head = output
 
+    # Avoid trying to do a test rebase if there are uncommitted changes.
+    # Either the rebase will fail with a similar message, or if the user
+    # has turned on rebase.autostash then the subsequent reset will
+    # silently discard those changes.
+    cmd = "git diff --quiet"
+    (status, output) = run_command_status(cmd)
+    if status != 0:
+        printwrap("You have unstaged changes. Please commit or stash them "
+                  "first, and then try again.")
+        sys.exit(1)
+    cmd = "git diff --cached --quiet"
+    (status, output) = run_command_status(cmd)
+    if status != 0:
+        printwrap("You have uncommitted changes. Please commit or stash them "
+                  "first, and then try again.")
+        sys.exit(1)
+
     cmd = "git show-ref --quiet --verify refs/%s" % remote_branch
     (status, output) = run_command_status(cmd)
     if status != 0:
@@ -935,67 +968,31 @@ def assert_one_change(remote, branch, yes, have_hook):
                       "branches (for independent changes).")
             print("\nThe outstanding commits are:\n\n%s\n\n"
                   "Do you really want to submit the above commits?" % output)
-            yes_no = do_input("Type 'yes' to confirm, other to cancel: ")
+            yes_no = input("Type 'yes' to confirm, other to cancel: ")
             if yes_no.lower().strip() != "yes":
                 print("Aborting.")
                 sys.exit(1)
 
 
-def use_topic(why, topic):
-    """Inform the user about why a particular topic has been selected."""
-    if VERBOSE:
-        print(why % ('"%s"' % topic,))
-    return topic
-
-
 def get_topic(target_branch):
-
     branch_name = get_branch_name(target_branch)
 
     branch_parts = branch_name.split("/")
     if len(branch_parts) >= 3 and branch_parts[0] == "review":
-        return use_topic("Using change number %s "
-                         "for the topic of the change submitted",
-                         "/".join(branch_parts[2:]))
+        # We don't want to set the review number as the topic
+        if branch_parts[2].isdigit():
+            return
 
-    preferred_log_format = "%B"
-    log_output = run_command("git log --pretty='" + preferred_log_format +
-                             "' HEAD^1..HEAD")
-    if log_output == preferred_log_format:
-        # The %B format specifier is supported starting at Git v1.7.2. If it's
-        # not supported, we'll just get back '%B', so we try something else.
-        # The downside of %s is that it removes newlines in the subject.
-        log_output = run_command("git log --pretty='%s%n%b' HEAD^1..HEAD")
-    bug_re = r'''(?x)                # verbose regexp
-                 \b([Bb]ug|[Ll][Pp]) # bug or lp
-                 [ \t\f\v]*          # don't want to match newline
-                 [:]?                # separator if needed
-                 [ \t\f\v]*          # don't want to match newline
-                 [#]?                # if needed
-                 [ \t\f\v]*          # don't want to match newline
-                 (\d+)               # bug number'''
+        topic = "/".join(branch_parts[2:])
+        if VERBOSE:
+            print("Using change number %s for the topic of the change "
+                  "submitted" % topic)
+        return topic
 
-    match = re.search(bug_re, log_output)
-    if match is not None:
-        return use_topic("Using bug number %s "
-                         "for the topic of the change submitted",
-                         "bug/%s" % match.group(2))
-
-    bp_re = r'''(?x)                         # verbose regexp
-                \b([Bb]lue[Pp]rint|[Bb][Pp]) # a blueprint or bp
-                [ \t\f\v]*                   # don't want to match newline
-                [#:]?                        # separator if needed
-                [ \t\f\v]*                   # don't want to match newline
-                ([0-9a-zA-Z-_]+)             # any identifier or number'''
-    match = re.search(bp_re, log_output)
-    if match is not None:
-        return use_topic("Using blueprint number %s "
-                         "for the topic of the change submitted",
-                         "bp/%s" % match.group(2))
-
-    return use_topic("Using local branch name %s "
-                     "for the topic of the change submitted",
-                     branch_name)
+    if VERBOSE:
+        print("Using local branch name %s for the topic of the change "
+              "submitted" % branch_name)
+    return branch_name
 
 
 class CannotQueryOpenChangesets(CommandFailed):
@@ -1152,8 +1149,8 @@ class CheckoutExistingBranchFailed(CommandFailed):
     EXIT_CODE = 65
 
 
-class ResetHardFailed(CommandFailed):
-    "Failed to hard reset downloaded branch"
+class ResetKeepFailed(CommandFailed):
+    "Failed to reset downloaded branch"
     EXIT_CODE = 66
 
 
@@ -1209,7 +1206,7 @@ def fetch_review(review, masterbranch, remote, project):
             refspec = review_info['currentPatchSet']['ref']
         else:
             refspec = [ps for ps in review_info['patchSets']
-                       if ps['number'] == patchset_number][0]['ref']
+                       if int(ps['number']) == int(patchset_number)][0]['ref']
     except IndexError:
         raise PatchSetNotFound(review_arg)
     except KeyError:
@@ -1222,7 +1219,8 @@ def fetch_review(review, masterbranch, remote, project):
     except KeyError:
         topic = review
     try:
-        author = re.sub('\W+', '_', review_info['owner']['name']).lower()
+        author = re.sub(r'\W+', '_',
+                        review_info['owner'][BRANCHAUTHOR]).lower()
     except KeyError:
         author = 'unknown'
     remote_branch = review_info['branch']
@@ -1234,7 +1232,7 @@ def fetch_review(review, masterbranch, remote, project):
 
     print("Downloading %s from gerrit" % refspec)
     run_command_exc(PatchSetGitFetchFailed,
-                    "git", "fetch", remote, refspec)
+                    "git", "fetch", remote_url, refspec)
     return branch_name, remote_branch
 
 
@@ -1248,13 +1246,14 @@ def checkout_review(branch_name, remote, remote_branch):
                         "git", "checkout", "-b",
                         branch_name, "FETCH_HEAD")
         # --set-upstream-to is supported starting in git 1.8
-        run_command_exc(SetUpstreamBranchFailed,
-                        "git", "branch", "--set-upstream-to",
-                        '%s/%s' % (remote, remote_branch),
-                        branch_name)
+        if remote is not None:
+            run_command_exc(SetUpstreamBranchFailed,
+                            "git", "branch", "--set-upstream-to",
+                            '%s/%s' % (remote, remote_branch),
+                            branch_name)
 
     except CheckoutNewBranchFailed as e:
-        if re.search("already exists\.?", e.output):
+        if re.search(r"already exists\.?", e.output):
             print("Branch %s already exists - reusing" % branch_name)
             track_remote, track_branch = parse_tracking(
                 ref='refs/heads/' + branch_name)
@@ -1267,8 +1266,8 @@ def checkout_review(branch_name, remote, remote_branch):
                 raise BranchTrackingMismatch
             run_command_exc(CheckoutExistingBranchFailed,
                             "git", "checkout", branch_name)
-            run_command_exc(ResetHardFailed,
-                            "git", "reset", "--hard", "FETCH_HEAD")
+            run_command_exc(ResetKeepFailed,
+                            "git", "reset", "--keep", "FETCH_HEAD")
         else:
             raise
 
@@ -1308,7 +1307,7 @@ class InvalidPatchsetsToCompare(GitReviewException):
     EXIT_CODE = 39
 
 
-def compare_review(review_spec, branch, remote, rebase=False):
+def compare_review(review_spec, branch, remote, project, rebase=False):
     new_ps = None    # none means latest
 
     if '-' in review_spec:
@@ -1321,7 +1320,7 @@ def compare_review(review_spec, branch, remote, rebase=False):
     old_review = build_review_number(review, old_ps)
     new_review = build_review_number(review, new_ps)
 
-    old_branch, _ = fetch_review(old_review, branch, remote)
+    old_branch, _ = fetch_review(old_review, branch, remote, project)
     checkout_review(old_branch, None, None)
 
     if rebase:
@@ -1331,7 +1330,11 @@ def compare_review(review_spec, branch, remote, rebase=False):
             print('Skipping rebase because of conflicts')
             run_command_exc(CommandFailed, 'git', 'rebase', '--abort')
 
-    new_branch, remote_branch = fetch_review(new_review, branch, remote)
+    new_branch, remote_branch = fetch_review(
+        new_review,
+        branch,
+        remote,
+        project)
     checkout_review(new_branch, remote, remote_branch)
 
     if rebase:
@@ -1377,18 +1380,72 @@ def assert_valid_reviewers(reviewers):
                 "Whitespace not allowed in reviewer: '%s'" % reviewer)
 
 
+class _DownloadFlag(argparse.Action):
+    """Special action for the various forms of downloading reviews.
+
+    Additional option parsing: store value in 'dest', but
+    at the same time set one of the flag options to True
+    """
+    def __call__(self, parser, namespace, value, option_string=None):
+        url = urlparse(value)
+        # Turn URLs into change ids:
+        #   https://review.opendev.org/423436
+        # and
+        #   https://review.opendev.org/423436/
+        # and
+        #   https://review.opendev.org/#/c/423436
+        # and
+        #   https://review.opendev.org/c/<project>/+/423436
+        # become
+        #   "423436"
+        # while
+        #   https://review.opendev.org/423436/1
+        # and
+        #   https://review.opendev.org/#/c/423436/1
+        # and
+        #   https://review.opendev.org/c/<project>/+/423436/1
+        # become
+        #   "423436,1".
+        #
+        # If there is a #, the rest of the path is stored in the
+        # "fragment", otherwise that will be empty.
+        base = url.fragment or url.path
+        parts = base.rstrip('/').lstrip('/c').split('/')
+        # PolyGerrit places the change after a '+' symbol in the url
+        try:
+            parts = parts[parts.index('+') + 1:]
+        except ValueError:
+            pass
+        change = parts[0]
+        if len(parts) > 1:
+            change = '%s,%s' % (change, parts[1])
+        setattr(namespace, self.dest, change)
+        setattr(namespace, self.const, True)
+
+
 def _main():
     usage = "git review [OPTIONS] ... [BRANCH]"
 
-    class DownloadFlag(argparse.Action):
-        """Additional option parsing: store value in 'dest', but
-           at the same time set one of the flag options to True
-        """
-        def __call__(self, parser, namespace, values, option_string=None):
-            setattr(namespace, self.dest, values)
-            setattr(namespace, self.const, True)
+    description = """\
+A git command for submitting branches to Gerrit.
 
-    parser = argparse.ArgumentParser(usage=usage, description=COPYRIGHT)
+git-review is a tool that helps submitting git branches to gerrit for
+review.
+"""
+
+    epilog = """\
+additional information:
+
+  Options configurable using: git config --global gitreview.OPTION VALUE
+
+  branchauthor          name (default), username, email
+"""
+
+    parser = argparse.ArgumentParser(
+        usage=usage,
+        description=description,
+        formatter_class=argparse.RawDescriptionHelpFormatter,
+        epilog=epilog)
 
     topic_arg_group = parser.add_mutually_exclusive_group()
     topic_arg_group.add_argument("-t", "--topic", dest="topic",
@@ -1399,13 +1456,6 @@ def _main():
 
     parser.add_argument("--reviewers", nargs="+",
                         help="Add reviewers to uploaded patch sets.")
-    parser.add_argument("-D", "--draft", dest="draft", action="store_true",
-                        help="Submit review as a draft")
-    parser.add_argument("-c", "--compatible", dest="compatible",
-                        action="store_true",
-                        help="Push change to refs/for/* for compatibility "
-                             "with Gerrit versions < 2.3. Ignored if "
-                             "-D/--draft is used.")
     parser.add_argument("-n", "--dry-run", dest="dry", action="store_true",
                         help="Don't actually submit the branch for review")
     parser.add_argument("-i", "--new-changeid", dest="regenerate",
@@ -1417,6 +1467,10 @@ def _main():
                         action="store_true",
                         help="Use remote push-url logic instead of separate"
                              " remotes")
+    parser.add_argument('--notify',
+                        choices=['NONE', 'OWNER', 'OWNER_REVIEWERS', 'ALL'],
+                        help="Control to whom email notifications are sent,"
+                             " defaults to ALL.")
 
     rebase_group = parser.add_mutually_exclusive_group()
     rebase_group.add_argument("-R", "--no-rebase", dest="rebase",
@@ -1438,31 +1492,34 @@ def _main():
     fetch.set_defaults(download=False, compare=False, cherrypickcommit=False,
                        cherrypickindicate=False, cherrypickonly=False)
     fetch.add_argument("-d", "--download", dest="changeidentifier",
-                       action=DownloadFlag, metavar="CHANGE",
+                       action=_DownloadFlag, metavar="CHANGE[,PS]",
                        const="download",
                        help="Download the contents of an existing gerrit "
-                            "review into a branch")
+                            "review into a branch. Include the patchset "
+                            "number to download a specific version of the "
+                            "change. The default is to take the most recent "
+                            "version.")
     fetch.add_argument("-x", "--cherrypick", dest="changeidentifier",
-                       action=DownloadFlag, metavar="CHANGE",
+                       action=_DownloadFlag, metavar="CHANGE",
                        const="cherrypickcommit",
                        help="Apply the contents of an existing gerrit "
                              "review onto the current branch and commit "
                              "(cherry pick; not recommended in most "
                              "situations)")
     fetch.add_argument("-X", "--cherrypickindicate", dest="changeidentifier",
-                       action=DownloadFlag, metavar="CHANGE",
+                       action=_DownloadFlag, metavar="CHANGE",
                        const="cherrypickindicate",
                        help="Apply the contents of an existing gerrit "
                        "review onto the current branch and commit, "
                        "indicating its origin")
     fetch.add_argument("-N", "--cherrypickonly", dest="changeidentifier",
-                       action=DownloadFlag, metavar="CHANGE",
+                       action=_DownloadFlag, metavar="CHANGE",
                        const="cherrypickonly",
                        help="Apply the contents of an existing gerrit "
                        "review to the working directory and prepare "
                        "for commit")
     fetch.add_argument("-m", "--compare", dest="changeidentifier",
-                       action=DownloadFlag, metavar="CHANGE,PS[-NEW_PS]",
+                       action=_DownloadFlag, metavar="CHANGE,PS[-NEW_PS]",
                        const="compare",
                        help="Download specified and latest (or NEW_PS) "
                        "patchsets of an existing gerrit review into "
@@ -1486,6 +1543,28 @@ def _main():
                              "you are submitting more than one patch")
     parser.add_argument("-v", "--verbose", dest="verbose", action="store_true",
                         help="Output more information about what's going on")
+
+    wip_group = parser.add_mutually_exclusive_group()
+    wip_group.add_argument("-w", "--work-in-progress", dest="wip",
+                           action="store_true",
+                           help="Send patch as work in progress for Gerrit "
+                                "versions >= 2.15")
+    wip_group.add_argument("-W", "--ready", dest="ready", action="store_true",
+                           help="Send patch that is already work in progress"
+                                " as ready for review. Gerrit versions >="
+                                " 2.15")
+
+    private_group = parser.add_mutually_exclusive_group()
+    private_group.add_argument("-p", "--private", dest="private",
+                               action="store_true",
+                               help="Send patch as a private patch ready for "
+                                    "review. Gerrit versions >= 2.15")
+    private_group.add_argument("-P", "--remove-private", dest="remove_private",
+                               action="store_true",
+                               help="Send patch which already in private state"
+                                    " to normal patch."" Gerrit versions >= "
+                                    "2.15")
+
     parser.add_argument("--no-custom-script", dest="custom_script",
                         action="store_false", default=True,
                         help="Do not run custom scripts.")
@@ -1509,12 +1588,16 @@ def _main():
     parser.add_argument("branch", nargs="?")
 
     parser.set_defaults(dry=False,
-                        draft=False,
                         verbose=False,
                         update=False,
                         setup=False,
                         list=False,
-                        yes=False)
+                        yes=False,
+                        wip=False,
+                        ready=False,
+                        private=False,
+                        remove_private=False)
+
     try:
         (top_dir, git_dir) = git_directories()
     except GitDirectoriesException as _no_git_dir:
@@ -1525,6 +1608,7 @@ def _main():
         parser.set_defaults(rebase=convert_bool(config['rebase']),
                             track=convert_bool(config['track']),
                             remote=None,
+                            notopic=convert_bool(config['notopic']),
                             usepushurl=convert_bool(config['usepushurl']))
     options = parser.parse_args()
 
@@ -1555,6 +1639,13 @@ def _main():
     yes = options.yes
     status = 0
 
+    global BRANCHAUTHOR
+    BRANCHAUTHOR = config['branchauthor']
+    if BRANCHAUTHOR not in ["name", "email", "username"]:
+        warn("Invalid gitreview.branchauthor: %s (using default: %s)"
+             % (BRANCHAUTHOR, DEFAULTS["branchauthor"]))
+        BRANCHAUTHOR = DEFAULTS["branchauthor"]
+
     if options.track:
         remote, branch = resolve_tracking(remote, branch)
 
@@ -1568,7 +1659,8 @@ def _main():
     if options.changeidentifier:
         if options.compare:
             compare_review(options.changeidentifier,
-                           branch, remote, options.rebase)
+                           branch, remote, config['project'],
+                           options.rebase)
             return
         local_branch, remote_branch = fetch_review(options.changeidentifier,
                                                    branch, remote,
@@ -1609,26 +1701,22 @@ def _main():
             sys.exit(1)
     assert_one_change(remote, branch, yes, have_hook)
 
-    ref = "publish"
+    ref = "for"
 
-    if options.draft:
-        ref = "drafts"
-        if options.custom_script:
-            run_custom_script("draft")
-    elif options.compatible:
-        ref = "for"
-
-    cmd = "git push %s HEAD:refs/%s/%s" % (remote, ref, branch)
+    cmd = ("git push --no-follow-tags %s HEAD:refs/%s/%s" %
+           (remote, ref, branch))
+    push_options = []
     if options.topic is not None:
         topic = options.topic
     else:
         topic = None if options.notopic else get_topic(branch)
+
     if topic and topic != branch:
-        cmd += "/%s" % topic
+        push_options.append("topic=%s" % topic)
 
     if options.reviewers:
         assert_valid_reviewers(options.reviewers)
-        cmd += "%" + ",".join("r=%s" % r for r in options.reviewers)
+        push_options += ["r=%s" % r for r in options.reviewers]
 
     if options.regenerate:
         print("Amending the commit to regenerate the change id\n")
@@ -1641,6 +1729,23 @@ def _main():
                         GIT_EDITOR="sed -i -e "
                         "'/^Change-Id:/d'")
 
+    if options.wip:
+        push_options.append('wip')
+
+    if options.ready:
+        push_options.append('ready')
+
+    if options.private:
+        push_options.append('private')
+
+    if options.remove_private:
+        push_options.append('remove-private')
+
+    if options.notify is not None:
+        push_options.append("notify=%s" % options.notify)
+
+    if push_options:
+        cmd += "%" + ",".join(push_options)
     if options.dry:
         print("Please use the following command "
               "to send your commits to review:\n")
@@ -1659,11 +1764,12 @@ def _main():
 
 
 def main():
+
     try:
         _main()
     except GitReviewException as e:
         print(e)
-        sys.exit(getattr(e, 'EXIT_CODE', -1))
+        sys.exit(e.EXIT_CODE)
 
 
 if __name__ == "__main__":
