@@ -18,6 +18,7 @@
 import json
 import os
 import shutil
+import testtools
 
 from git_review import tests
 from git_review.tests import utils
@@ -143,6 +144,21 @@ class GitReviewTestCase(tests.BaseGitReviewTestCase):
             'refs/remotes/%s/master' % self._remote,
             self._run_git("for-each-ref", "--format='%(upstream)'", head))
 
+        # add some more changes & upload
+        self._simple_amend('test 2nd rev',
+                           self._dir('test', '2nd_rev_file.txt'))
+        self._run_git_review('-v')
+        self._simple_amend('test 3rd rev',
+                           self._dir('test', '3rd_rev_file.txt'))
+        self._run_git_review('-v')
+
+        # get rev 2; assert rev2 file is there, but not rev3
+        self._run_git_review('-d', '%s,%s' % (change_id, 2))
+        self.assertIn('2nd_rev_file.txt',
+                      self._run_git('show', 'HEAD'))
+        self.assertNotIn('3rd_rev_file.txt',
+                         self._run_git('show', 'HEAD'))
+
     def test_multiple_changes(self):
         """Test git-review asks about multiple changes.
 
@@ -206,6 +222,26 @@ class GitReviewTestCase(tests.BaseGitReviewTestCase):
 
         self.assertEqual(set(['reviewer1', 'reviewer2']), reviewers)
 
+    def test_rebase_unstaged_changes_msg(self):
+        """Test message displayed when unstaged changes are present."""
+        self._run_git_review('-s')
+        self._run_git('checkout', '-b', 'test_branch')
+        # By not passing a filename, we rely on an edit to a preprovisioned
+        # tracked default file, which will be unstaged
+        self._unstaged_change(change_text='simple message')
+        exc = self.assertRaises(Exception, self._run_git_review)
+        self.assertIn("You have unstaged changes. Please", exc.args[0])
+
+    def test_rebase_uncommitted_changes_msg(self):
+        """Test message displayed when staged changes are present."""
+        self._run_git_review('-s')
+        self._run_git('checkout', '-b', 'test_branch')
+        # By passing a filename, we rely on an edit to a new tracked
+        # file, which will be staged
+        self._uncommitted_change(change_text='simple message')
+        exc = self.assertRaises(Exception, self._run_git_review)
+        self.assertIn("You have uncommitted changes. Please", exc.args[0])
+
     def test_rebase_no_remote_branch_msg(self):
         """Test message displayed where no remote branch exists."""
         self._run_git_review('-s')
@@ -271,9 +307,10 @@ class GitReviewTestCase(tests.BaseGitReviewTestCase):
 
         br_out = self._run_git('checkout',
                                '-b', 'test_branch', 'origin/maint')
-        expected_track = 'Branch test_branch set up to track remote' + \
-                         ' branch maint from origin.'
-        self.assertIn(expected_track, br_out)
+        expected_track = (".*\nBranch '?test_branch'? set up to track remote "
+                          "branch '?maint'? from '?origin'?.")
+        track_matcher = testtools.matchers.MatchesRegex(expected_track)
+        self.assertThat(br_out, track_matcher)
         branches = self._run_git('branch', '-a')
         expected_branch = '* test_branch'
         observed = branches.split('\n')
@@ -368,99 +405,42 @@ class GitReviewTestCase(tests.BaseGitReviewTestCase):
         finally:
             os.environ.update(LANG=lang_env)
 
+    def test_no_topic(self):
+        """Test on change with no topic.
+
+        This will be checked out as 'review/{owner}/{ID}'. We don't want to set
+        the topic to '{ID}'.
+        """
+        self._run_git_review('-s')
+        curr_branch = self._run_git('rev-parse', '--abbrev-ref', 'HEAD')
+        self._run_git('checkout', '-b', 'review/johndoe/123456')
+        self._simple_change('test file modified', 'derp derp derp')
+        self._assert_branch_would_be(curr_branch)
+
     def test_git_review_t(self):
         self._run_git_review('-s')
         self._simple_change('test file modified', 'commit message for bug 654')
-        self._assert_branch_would_be('master/zat', extra_args=['-t', 'zat'])
+        self._assert_branch_would_be('master%topic=zat',
+                                     extra_args=['-t', 'zat'])
 
-    def test_bug_topic(self):
-        self._run_git_review('-s')
-        self._simple_change('a change', 'new change for bug 123')
-        self._assert_branch_would_be('master/bug/123')
-
-    def test_bug_topic_newline(self):
-        self._run_git_review('-s')
-        self._simple_change('a change', 'new change not for bug\n\n123')
-        self._assert_branch_would_be('master')
-
-    def test_bp_topic(self):
-        self._run_git_review('-s')
-        self._simple_change('a change', 'new change for blueprint asdf')
-        self._assert_branch_would_be('master/bp/asdf')
-
-    def test_bp_topic_newline(self):
-        self._run_git_review('-s')
-        self._simple_change('a change', 'new change not for blueprint\n\nasdf')
-        self._assert_branch_would_be('master')
-
-    def test_author_name_topic_bp(self):
-        old_author = None
-        if 'GIT_AUTHOR_NAME' in os.environ:
-            old_author = os.environ['GIT_AUTHOR_NAME']
-        try:
-            os.environ['GIT_AUTHOR_NAME'] = 'BPNAME'
-            self._run_git_review('-s')
-            self._simple_change('a change',
-                                'new change 1 with name but no topic')
-            self._assert_branch_would_be('master')
-        finally:
-            if old_author:
-                os.environ['GIT_AUTHOR_NAME'] = old_author
-            else:
-                del os.environ['GIT_AUTHOR_NAME']
-
-    def test_author_email_topic_bp(self):
-        old_author = None
-        if 'GIT_AUTHOR_EMAIL' in os.environ:
-            old_author = os.environ['GIT_AUTHOR_EMAIL']
-        try:
-            os.environ['GIT_AUTHOR_EMAIL'] = 'bpemail@example.com'
-            self._run_git_review('-s')
-            self._simple_change('a change',
-                                'new change 1 with email but no topic')
-            self._assert_branch_would_be('master')
-        finally:
-            if old_author:
-                os.environ['GIT_AUTHOR_EMAIL'] = old_author
-            else:
-                del os.environ['GIT_AUTHOR_EMAIL']
-
-    def test_author_name_topic_bug(self):
-        old_author = None
-        if 'GIT_AUTHOR_NAME' in os.environ:
-            old_author = os.environ['GIT_AUTHOR_NAME']
-        try:
-            os.environ['GIT_AUTHOR_NAME'] = 'Bug: #1234'
-            self._run_git_review('-s')
-            self._simple_change('a change',
-                                'new change 2 with name but no topic')
-            self._assert_branch_would_be('master')
-        finally:
-            if old_author:
-                os.environ['GIT_AUTHOR_NAME'] = old_author
-            else:
-                del os.environ['GIT_AUTHOR_NAME']
-
-    def test_author_email_topic_bug(self):
-        old_author = None
-        if 'GIT_AUTHOR_EMAIL' in os.environ:
-            old_author = os.environ['GIT_AUTHOR_EMAIL']
-        try:
-            os.environ['GIT_AUTHOR_EMAIL'] = 'bug5678@example.com'
-            self._run_git_review('-s')
-            self._simple_change('a change',
-                                'new change 2 with email but no topic')
-            self._assert_branch_would_be('master')
-        finally:
-            if old_author:
-                os.environ['GIT_AUTHOR_EMAIL'] = old_author
-            else:
-                del os.environ['GIT_AUTHOR_EMAIL']
+        # -t takes precedence over notopic
+        self._run_git('config', 'gitreview.notopic', 'true')
+        self._assert_branch_would_be('master%topic=zat',
+                                     extra_args=['-t', 'zat'])
 
     def test_git_review_T(self):
         self._run_git_review('-s')
+        self._run_git('checkout', '-b', 'bug/456')
         self._simple_change('test file modified', 'commit message for bug 456')
-        self._assert_branch_would_be('master/bug/456')
+        self._assert_branch_would_be('master%topic=bug/456')
+        self._assert_branch_would_be('master', extra_args=['-T'])
+
+        self._run_git('config', 'gitreview.notopic', 'true')
+        self._assert_branch_would_be('master')
+        self._run_git('config', 'gitreview.notopic', 'false')
+        self._assert_branch_would_be('master%topic=bug/456')
+
+        # -T takes precedence over notopic=false
         self._assert_branch_would_be('master', extra_args=['-T'])
 
     def test_git_review_T_t(self):
@@ -480,8 +460,8 @@ class GitReviewTestCase(tests.BaseGitReviewTestCase):
         self._run_gerrit_cli('review', head, '--abandon')
 
         # Populate "project2" repo
-        self._run_gerrit_cli('create-project', '--empty-commit', '--name',
-                             'test/test_project2')
+        self._run_gerrit_cli('create-project', 'test/test_project2',
+                             '--empty-commit')
         project2_uri = self.project_uri.replace('test/test_project',
                                                 'test/test_project2')
         self._run_git('fetch', project2_uri, 'HEAD')
